<?php

/**
 * Music Library Manager - Song Editor 
 * 
 * This page is used to add new songs to the DB or edit existing ones. Because
 * the presentation and user experience of both is very similar, they share the
 * same page. 
 * 
 * Note that this page mostly contains presentation and usability logic and
 * uses backend classes for business logic and data management.
 * 
 * If invalid data is passed to the model layer, it is thrown back as an 
 * exception with a list of detailed errors. These errors are presented back to
 * the user in this page. 
 * 
 * If a song is added or updated successfuly, the user is redirected back to 
 * the home page.  
 */

// Load application configuration
require_once '../config/jukebox-config.php';

// Load model class
require_once '../library/JukeboxSong.php';

$error = null;
$fieldErrors = array();

// Create an empty song object for convenience
$song = new JukeboxSong(); 

// Is this a POST or GET request?
if ($_SERVER['REQUEST_METHOD'] == 'POST') {
    // If POST, handle form submission
    
    try {
        $song = new JukeboxSong($_POST);
        
        // If we have a song ID - set id (this means we are saving changes to
        // an existing song
        if (isset($_GET['songId'])) {
            $song->id = $_GET['songId'];
        }

        // Save changes
        $song->save();
    
        // Redirect back to the listing page
        header("Location: index.php");
        exit;
        
    // Catch data validation errors
    } catch (JukeboxDataException $ex) {
        $fieldErrors = $ex->getAllFieldMessages();
        $error = $ex->getMessage();
        
    // Catch all other errors (e.g. PDO errors)
    } catch (Exception $ex) {
        $error = "Error saving song: " . $ex->getMessage();
    }
}

// Are we editing or adding?
if (isset($_GET['songId'])) {
    $song = JukeboxSong::fetchSong($_GET['songId']);
    $pageTitle = "Edit Song: $song->title";
} else {
    $pageTitle = 'Add Song';
}

$genres = JukeboxSong::getGenresList();

// HTML header
require_once 'common/html-top.php';
?>
	<h2><?php echo $pageTitle ?></h2>
<?php if ($error): ?>
	<div class="error">
		<?php echo htmlspecialchars($error) ?>
	</div>
<?php endif; ?>	
	<form method="POST">
		<dl>
			<dt><label for="title">Song Title:</label></dt>
			<dd><input type="text" id="title" name="title" maxlength="72" value="<?php echo htmlspecialchars($song->title) ?>" /></dd>
<?php if (isset($fieldErrors['title'])): ?>
			<dd class="fieldError"><?php echo htmlspecialchars($fieldErrors['title']); ?></dd>
<?php endif; ?>
			
			<dt><label for="artist">Artist:</label></dt>
			<dd><input type="text" id="artist" name="artist" maxlength="72" value="<?php echo htmlspecialchars($song->artist) ?>" /></dd>
<?php if (isset($fieldErrors['artist'])): ?>
			<dd class="fieldError"><?php echo htmlspecialchars($fieldErrors['artist']); ?></dd>
<?php endif; ?>
			
			<dt><label for="year">Year:</label></dt>
			<dd><input type="text" id="year" name="year" maxlength="4" value="<?php echo htmlspecialchars($song->year) ?>" /></dd>
<?php if (isset($fieldErrors['year'])): ?>
			<dd class="fieldError"><?php echo htmlspecialchars($fieldErrors['year']); ?></dd>
<?php endif; ?>
			
			<dt><label for="genre">Genre:</label></dt>
			<dd><select name="genre">
<?php foreach($genres as $genreId => $genre): 
    $slctd = ($song->genre == $genreId);
?>
				<option value="<?php echo $genreId; if ($slctd) echo '" selected="selected'; ?>"><?php echo htmlspecialchars($genre) ?></option> 
<?php endforeach; ?>
			</select></dd>
<?php if (isset($fieldErrors['genre'])): ?>
			<dd class="fieldError"><?php echo htmlspecialchars($fieldErrors['genre']); ?></dd>
<?php endif; ?>
			
			<dt>&nbsp;</dt>
			<dd><input type="submit" value="Save" />
				<input type="button" value="Cancel" onclick="window.location='index.php'" />
			</dd>
		</dl>
	</form>
	
<?php 

require_once 'common/html-bottom.php';