<?php

/**
 * Music Library Management App - Song class
 * 
 * This class contains most of the model / business logic of the application. 
 * 
 * It mostly represents a song object and provides access to the song's 
 * properties through 'magic' __get() and __set() methods (methods which, if
 * exist, are called on read/write of object properties). 
 * 
 * Additionally, this class contains a set of static functions to access the 
 * list of all songs as well as the list of genres.
 * 
 * Note that the actual database schema is defined in data/schema.sql 
 */

require_once 'JukeboxDb.php';

class JukeboxSong
{
    protected $_data = array(
        'id'     => null,
        'title'  => null,
        'artist' => null,
        'year'   => null,
        'genre'  => null
    ); 
    
    /**
     * Create a new song object from array of data
     *
     * @param array $data
     */
    public function __construct(array $data = array())
    {
        foreach($data as $key => $value) {
            if (array_key_exists($key, $this->_data)) {
                $this->_data[$key] = $value; 
            }
        }
    }
    
    /**
     * Ensure that we have all the data we need in order to create/update a song
     */
    
    
    private function validate()
    {
        $title  = null;
        $artist = null;
        $year   = null;
        $genre  = null;
        $errors = array();

    	// Validate title
        if (isset($this->_data['title'])) {
            $title = trim($this->_data['title']);
            if (! $title) {
                $errors['title'] = 'Title must be specified';
                
            } elseif (strlen($title) > 72) {
                $errors['title'] = 'Title must be up to 72 characters long';
            }
        } else {
            $errors['title'] = 'Title must be specified';
        }
        
        // Validate artist
        if (isset($this->_data['artist'])) {
            $artist = trim($this->_data['artist']);
            if (! $artist) {
                $errors['artist'] = 'Artist must be specified';
                
            } elseif (strlen($artist) > 72) {
                $errors['artist'] = 'Artist must be up to 72 characters long';
                
            }
            
        } else {
            $errors['artist'] = 'Artist must be specified';
        }
        
        // Validate year
        if (isset($this->_data['year'])) {
            $year = (int) trim($this->_data['year']);
            $this_year = (int) date('Y');
            
            // Year can be between 1000 and the current year
            if ($year < 1000 || $year > $this_year) {
                $errors['year'] = 'Year must be specified as a number between 1000 and ' . $this_year;
            } 
        } else {
            $errors['year'] = 'You must specify the song\'s publish year';
        }
        
        // Validate Genre
        if (isset($this->_data['genre'])) {
            // Force integer for $genre
            $genre = (int) $this->_data['genre'];
            
            // Make sure Genre is listed in DB
            
            // Query DB for this genre. This would not be required if we were 
            // using a foreign key constraint. 
            // No data escaping is needed because $genre is always integer
            $result = JukeboxDb::getConnection()->query("SELECT 1 FROM genres WHERE id = $genre");
            if (! $result->fetch()) {
                $errors['genre'] = 'Invalid genre selected';
            }
        } else {
            $errors['genre'] = 'Genre must be selected';
        }
        
        // Do we have any errors? If so, throw an exception that will be caught
        // in our calling script
        if (! empty($errors)) {
            require_once 'JukeboxDataException.php';
            $ex = new JukeboxDataException("Missing or invalid song data");
            foreach($errors as $field => $message) {
                $ex->setFieldMessage($field, $message);
            }
            
            throw $ex;
        }
        
        return array(
        	'title'		=> $title,
        	'artist'	=> $artist,
        	'year'		=> $year,
        	'genre'		=> $genre);
    }
    
	/**
     * Save song as a new song (SQL INSERT) or as an existing one (SQL UPDATE)
     *
     * @param array   $data
     * @param integer $songId
     */
    public function save()
    {
        $db = JukeboxDb::getConnection();
        

        // Ensure we have all the data we need and that it's properly formatted.
        // In case we don't - an exception will be thrown, and we let it propagate.
        $data = $this->validate();
        
        // Are we updating or inserting
        if ($this->_data['id'] === null) {
            // If we are inserting, we need to find the highest song ID and 
            // increment it. If we were using the DB auto-increment feature
            // this would not be required.

            $result = $db->query("SELECT MAX(id) FROM music");
            if (! ($maxId = $result->fetchColumn(0))) {
                $maxId = 0;
            }
            
            $result = null;

            $stmt = $db->prepare(
            	"INSERT INTO music (id, title, artist, year, genre) " . 
                "  VALUES (?, ?, ?, ?, ?)"
            );

            // Since we have no auto-increment, it is possible for two users
            // to try to write using the same song ID, creating a race 
            // condition. For this reason, if we fail due to a UNIQUE constraint,
            // we try to save over again.
            // This relies on the ID field being defined as a UNIQUE INDEX in the DB. 
            while (true) {
                try {
                    $stmt->execute(array(++$maxId, 
                                         $data['title'],
                                         $data['artist'],
                                         $data['year'],
                                         $data['genre']));

                    break;
                    
                } catch (PDOException $ex) {
                    // If error is not a UNIQUE constraint violation, throw exception
                    // Otherwise try increasing the ID by 1.
                    if ($ex->getCode() != 23000) throw $ex;
                }
            }
        } else {
            // We are updating an existing song
            $stmt = $db->prepare(
            	"UPDATE music " . 
    			"  SET title = ?, artist = ?, year = ?, genre = ? " .  
            	"  WHERE id = ?"
            );
            
            // We do not check if songId actually exist. The assumption is that
            // if it doesn't, it can be silently ignored because the user 
            // probably tried something funny.
            $stmt->execute(array($data['title'], $data['artist'], $data['year'], $data['genre'], $this->_data['id']));
        }
    }
    
    /**
     * Magic method to access song properties
     *
     * @param  string $field
     * @return mixed
     */
    public function __get($field)
    {
        if (isset($this->_data[$field])) {
            return $this->_data[$field];
        } else {
            return null;
        }
    }
    
    /**
     * Magic method to mutate song properties
     *
     * @param string $field
     * @param string $value
     */
    public function __set($field, $value)
    {
        if (array_key_exists($field, $this->_data)) {
            $this->_data[$field] = $value;
        } else {
            throw new ErrorException("No such field '$field' in Song objects");
        }
    }
    
    /*****************
     * Static Methods
     ****************/
    
    /**
     * Get an associative array representing the list of genres
     * 
     * @todo   This can be optimized through caching
     * 
     * @return array
     */
    public static function getGenresList()
    {
        $db = JukeboxDb::getConnection();
        
        $query = "SELECT id, desc FROM genres ORDER BY desc";
        
        $stmt = $db->prepare($query);
        $stmt->execute();
        
        $stmt->bindColumn('id', $gen_id, PDO::PARAM_INT);
        $stmt->bindColumn('desc', $gen_title, PDO::PARAM_STR);
        
        $genres = array();
        while($stmt->fetch(PDO::FETCH_BOUND)) {
            $genres[$gen_id] = $gen_title;
        }
        
        return $genres;
    }
    
    /**
     * Get the list of all songs in DB, optionally using a filter, paginated
     *
     * @param  array   $filter
     * @param  integer $page
     * @return array
     */
    public static function getSongList(array $filters = array(), $searchSubstrings=true)
    {
        $db = JukeboxDb::getConnection();
        
        $query = "SELECT music.id AS id, title, artist, year, genre, desc AS gen_desc " . 
                 "FROM music LEFT JOIN genres ON music.genre = genres.id ";
        
        // Trim and sanitize filters
        foreach($filters as $key => $value) {
            $value = trim($value);
            if ($value === '') {
                unset($filters[$key]);
            } else {
                $filters[$key] = $value;
            }
        }
        
        // Add filters to query
        $where = array();

        // Should we use LIKE and %%'s, or a strict equality?       
        if ($searchSubstrings) {
        	$comparisonOperator = "LIKE ";
        	$prefixSuffix = '%';
        } else {
        	$comparisonOperator = "= ";
        	$prefixSuffix = '';
        }
        
        if (isset($filters['title'])) {
        	$where[] = 'title ' . $comparisonOperator . 
        		$db->quote($prefixSuffix . $filters['title'] . $prefixSuffix, PDO::PARAM_STR);
        }
        
        if (isset($filters['artist'])) {
        	$where[] = 'artist ' . $comparisonOperator . 
        		$db->quote($prefixSuffix . $filters['artist'] . $prefixSuffix, PDO::PARAM_STR);
        }
        
        if (isset($filters['genre'])) {
            $where[] = "genre = " . 
                $db->quote($filters['genre'], PDO::PARAM_INT);
        }
        
        if (isset($filters['year'])) {
            $where[] = "year = " . 
                $db->quote($filters['year'], PDO::PARAM_INT);
        }
        
        if (! empty($where)) {
            $query .= " WHERE " . implode(" AND ", $where);
        }
        
        $query .= "ORDER BY artist, title";
        
        $result = $db->query($query);
        $data = array();
        
        while (($row = $result->fetch(PDO::FETCH_ASSOC))) {
            $data[] = $row;
        }
        
        return $data;
    }
    
    /**
     * Get a particular song from the DB according to it's ID
     *
     * @param  integer $songId
     * @return JukeboxSong | null
     */
    public static function fetchSong($songId)
    {
        $db = JukeboxDb::getConnection();
        
        // Sanitize $songId (must be integer)
        $songId = (int) $songId;
        
        $query = "SELECT id, title, artist, year, genre FROM music " . 
                 "  WHERE id = $songId";
        
        $result = $db->query($query);
        if (($row = $result->fetch(PDO::FETCH_ASSOC))) {
            $song = new JukeboxSong($row);
            return $song;
        } else {
            return null;
        }
    }
}